/*
# $BEGIN_KEYHOLE_LICENSE$
# 
# This file is part of the Keyhole project, a lightweight library for
# the Arduino IDE, for interpreting commands and communicating variable
# values via a JSON- and Python-compatible text interface. 
# 
# Copyright (c) 2023- Jeremy Hill
# 
# Behold the 3-Clause BSD License under which it is hereby released:
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
# 
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# 
# $END_KEYHOLE_LICENSE$

*/
#ifndef   __Keyhole_H__
#define   __Keyhole_H__

#include "Arduino.h" // for Stream and String
/*
  The following example allows two global variables to be queried and/or
  set over a serial port or other Stream::
  
      #include "Keyhole.h"
    
      double foo = 0.0;
      String bar = "look at these escaped characters: \" \x08 \\ \t\r\n";
    
      void setup()
      {
        Serial.begin(9600);
      }
    
      void loop()
      {
        // main business of your sketch goes here
        // ...
        
        KEYHOLE keyhole(Serial);
        if( keyhole.begin() )
        {
          if(keyhole.variable("foo", foo)) Serial.println("assigned to foo");
          if(keyhole.variable("bar", bar)) Serial.println("assigned to bar");
          if(keyhole.command("Marco!")) Serial.println("Polo!");
          keyhole.end();
        }
      }
      
  An optional third argument to `variable()` can be `VARIABLE_READ_ONLY`,
  `VARIABLE_VERBOSE`, or the default value `VARIABLE_SILENT`.
  
  In this example, if you send the command `foo` over the serial port, the
  reply will be a JSON string associating the name `foo` with the current
  value of the corresponding global variable `gFoo`. 
  
  Provided the variable is not designated as `VARIABLE_READ_ONLY`, you can
  also set its value, e.g. by sending the command `foo = 2`.  (If you have
  designated the variable as `VARIABLE_VERBOSE` then this will also trigger
  a JSON output, as if you had queried `foo` immediately after setting it.)

  You can send the simple command `?` to receive a JSON output containing all
  the variables that are accessible in this manner.

  You can also have the full report delivered automatically on a repeating
  schedule by setting the `.autoSeconds` member greater than zero; you can
  even allow this parameter itself to be controlled via the keyhole, by
  exposing the `autoSeconds` member itself as a variable::

        KEYHOLE keyhole(Serial);
        if(keyhole.begin())
        {
          if(keyhole.variable("foo", foo)) Serial.println("assigned to foo");
          if(keyhole.variable("bar", bar)) Serial.println("assigned to bar");
          keyhole.variable("report_period_sec", keyhole.autoSeconds);
          keyhole.end();
        }

  Under the hood: `keyhole` is an instance of the class `Keyhole`. Such
  instances must either be global variables, or declared `static`. The
  `KEYHOLE` macro can be used inside or outside of `loop()` - it simply
  declares a `static Keyhole` instance (while hiding the possibly-not-so-
  user-friendly keyword `static` from users who are not familiar with it).
  
*/

// Debugging macros:
#define DBSTREAM Serial
#define REPORT( X )  { DBSTREAM.print( "{\"" #X "\" : " ); DBSTREAM.print( X );                         DBSTREAM.println( "}" ); DBSTREAM.flush(); }
#define REPORTS( X ) { DBSTREAM.print( "{\"" #X "\" : " ); Keyhole( DBSTREAM ).printStringLiteral( X ); DBSTREAM.println( "}" ); DBSTREAM.flush(); }

typedef enum
{
	VARIABLE_READ_ONLY = 0,
	VARIABLE_SILENT    = 1,
	VARIABLE_VERBOSE   = 2
} VariableWriteMode;

#define KEYHOLE       static Keyhole
class Keyhole
{
	public:
		Keyhole( Stream & stream=Serial, float autoSeconds=0.0, bool plotterMode=false );
		~Keyhole();
	
		bool begin( void );
		bool begin( unsigned long microsecondTimestamp );
	
		bool command( const char * cmd );
	
		bool variable( const char * variableName, bool &           referenceToVariable,  VariableWriteMode mode=VARIABLE_SILENT );
		bool variable( const char * variableName, char &           referenceToVariable,  VariableWriteMode mode=VARIABLE_SILENT );
		bool variable( const char * variableName, int8_t &         referenceToVariable,  VariableWriteMode mode=VARIABLE_SILENT );
		bool variable( const char * variableName, unsigned char &  referenceToVariable,  VariableWriteMode mode=VARIABLE_SILENT );
		bool variable( const char * variableName, int &            referenceToVariable,  VariableWriteMode mode=VARIABLE_SILENT );
		bool variable( const char * variableName, unsigned int &   referenceToVariable,  VariableWriteMode mode=VARIABLE_SILENT );
		bool variable( const char * variableName, short &          referenceToVariable,  VariableWriteMode mode=VARIABLE_SILENT );
		bool variable( const char * variableName, unsigned short & referenceToVariable,  VariableWriteMode mode=VARIABLE_SILENT );
		bool variable( const char * variableName, long &           referenceToVariable,  VariableWriteMode mode=VARIABLE_SILENT );
		bool variable( const char * variableName, unsigned long &  referenceToVariable,  VariableWriteMode mode=VARIABLE_SILENT );
		bool variable( const char * variableName, float &          referenceToVariable,  VariableWriteMode mode=VARIABLE_SILENT );
		bool variable( const char * variableName, double &         referenceToVariable,  VariableWriteMode mode=VARIABLE_SILENT );
		bool variable( const char * variableName, String &         referenceToVariable,  VariableWriteMode mode=VARIABLE_SILENT );
#		define variableAssigned variable
	
		bool end( void );
	
		unsigned long elapsed( void );
		void          printStringLiteral( const String & s, char withQuotes='"' );
		void          printCharLiteral(   char c,           char withQuotes=(char)-1 );
		// withQuotes=0 means print the character (or escape code) unquoted; -1 causes the
		// char to be printed as a numeric int8_t or uint8_t (depending on architecture) -
		// this is the default for reasons discussed in the comments of the method.

		Stream &      stream;
		float         autoSeconds;
		int           plotterMode;
		
	private:
		unsigned long mBeginMicros;
		String        mFullCommand;
		int           mListAllVariables;
		String        mPartialCommand;
		bool          mBackslash;
		int           mHexEscape;
		char          mHexValue;
		char          mQuote;
		unsigned long mTimestampOfLastAutoReport;
		char          mBad;

		const char *  _parseVariableCommand( const char * variableName, unsigned int & commandLength );
};

unsigned long StrToUnsignedInteger( const char * start, char ** endptr );
long          StrToSignedInteger(   const char * start, char ** endptr );
double        StrToDouble( const char * start, char ** endptr );

#endif // __Keyhole_H__
